"""WSGI SAML package for SAML 2.0 Attribute and Authorisation Decision Query/
Request Profile interfaces

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "15/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: queryinterface.py 7526 2010-09-27 12:08:12Z pjkersha $"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
import logging
log = logging.getLogger(__name__)
import traceback
from cStringIO import StringIO
from uuid import uuid4
from datetime import datetime, timedelta

from ndg.soap.server.wsgi.middleware import SOAPMiddleware
from ndg.soap.etree import SOAPEnvelope

from ndg.saml.utils import str2Bool
from ndg.saml.utils.factory import importModuleObject
from ndg.saml.xml import UnknownAttrProfile
from ndg.saml.common import SAMLVersion
from ndg.saml.utils import SAMLDateTime
from ndg.saml.saml2.core import (Response, Status, StatusCode, StatusMessage, 
                                 Issuer) 
from ndg.saml.saml2.binding.soap import SOAPBindingInvalidResponse


class SOAPQueryInterfaceMiddlewareError(Exception):
    """Base class for WSGI SAML 2.0 SOAP Query Interface Errors"""


class SOAPQueryInterfaceMiddlewareConfigError(Exception):
    """WSGI SAML 2.0 SOAP Query Interface Configuration problem"""


class QueryIssueInstantInvalid(SOAPBindingInvalidResponse):
    """Invalid timestamp for incoming query"""
    
    
class SOAPQueryInterfaceMiddleware(SOAPMiddleware):
    """Implementation of SAML 2.0 SOAP Binding for Query/Request Binding
    
    @type PATH_OPTNAME: basestring
    @cvar PATH_OPTNAME: name of app_conf option for specifying a path or paths
    that this middleware will intercept and process
    @type QUERY_INTERFACE_KEYNAME_OPTNAME: basestring
    @cvar QUERY_INTERFACE_KEYNAME_OPTNAME: app_conf option name for key name
    used to reference the SAML query interface in environ
    @type DEFAULT_QUERY_INTERFACE_KEYNAME: basestring
    @param DEFAULT_QUERY_INTERFACE_KEYNAME: default key name for referencing
    SAML query interface in environ
    """
    log = logging.getLogger('SOAPQueryInterfaceMiddleware')
    PATH_OPTNAME = "mountPath"
    QUERY_INTERFACE_KEYNAME_OPTNAME = "queryInterfaceKeyName"
    DEFAULT_QUERY_INTERFACE_KEYNAME = ("ndg.security.server.wsgi.saml."
                            "SOAPQueryInterfaceMiddleware.queryInterface")
    
    REQUEST_ENVELOPE_CLASS_OPTNAME = 'requestEnvelopeClass'
    RESPONSE_ENVELOPE_CLASS_OPTNAME = 'responseEnvelopeClass'
    SERIALISE_OPTNAME = 'serialise'
    DESERIALISE_OPTNAME = 'deserialise' 
    SAML_VERSION_OPTNAME = 'samlVersion'
    ISSUER_NAME_OPTNAME = 'issuerName'
    ISSUER_FORMAT_OPTNAME = 'issuerFormat'
    
    CONFIG_FILE_OPTNAMES = (
        PATH_OPTNAME,
        QUERY_INTERFACE_KEYNAME_OPTNAME,
        DEFAULT_QUERY_INTERFACE_KEYNAME,
        REQUEST_ENVELOPE_CLASS_OPTNAME,
        RESPONSE_ENVELOPE_CLASS_OPTNAME,
        SERIALISE_OPTNAME,
        DESERIALISE_OPTNAME,
        SAML_VERSION_OPTNAME,
        ISSUER_NAME_OPTNAME,
        ISSUER_FORMAT_OPTNAME
    )
    
    def __init__(self, app):
        '''@type app: callable following WSGI interface
        @param app: next middleware application in the chain 
        '''     
        super(SOAPQueryInterfaceMiddleware, self).__init__()
        
        self._app = app
        
        # Set defaults
        cls = SOAPQueryInterfaceMiddleware
        self.__queryInterfaceKeyName = cls.DEFAULT_QUERY_INTERFACE_KEYNAME
        self.__mountPath = None
        self.mountPath = '/'
        self.__requestEnvelopeClass = None
        self.__responseEnvelopeClass = None
        self.__serialise = None
        self.__deserialise = None
        self.__issuer = None
        self.__clockSkewTolerance = timedelta(seconds=0.)
        self.__verifyTimeConditions = True
        self.__verifySAMLVersion = True
        self.__samlVersion = SAMLVersion.VERSION_20
        
        # Proxy object for SAML Response Issuer attributes.  By generating a 
        # proxy the Response objects inherent attribute validation can be 
        # applied to Issuer related config parameters before they're assigned to
        # the response issuer object generated in the authorisation decision 
        # query response
        self.__issuerProxy = Issuer()
      
    def initialise(self, global_conf, prefix='', **app_conf):
        '''
        @type global_conf: dict        
        @param global_conf: PasteDeploy global configuration dictionary
        @type prefix: basestring
        @param prefix: prefix for configuration items
        @type app_conf: dict        
        @param app_conf: PasteDeploy application specific configuration 
        dictionary
        '''
        # Override where set in config
        for name in SOAPQueryInterfaceMiddleware.CONFIG_FILE_OPTNAMES:
            val = app_conf.get(prefix + name)
            if val is not None:
                setattr(self, name, val)

        if self.serialise is None:
            raise AttributeError('No "serialise" method set to serialise the '
                                 'SAML response from this middleware.')

        if self.deserialise is None:
            raise AttributeError('No "deserialise" method set to parse the '
                                 'SAML request to this middleware.')
            
    def _getSerialise(self):
        return self.__serialise

    def _setSerialise(self, value):
        if isinstance(value, basestring):
            self.__serialise = importModuleObject(value)
            
        elif callable(value):
            self.__serialise = value
        else:
            raise TypeError('Expecting callable for "serialise"; got %r' % 
                            value)

    serialise = property(_getSerialise, _setSerialise, 
                         doc="callable to serialise request into XML type")

    def _getDeserialise(self):
        return self.__deserialise

    def _setDeserialise(self, value):
        if isinstance(value, basestring):
            self.__deserialise = importModuleObject(value)
            
        elif callable(value):
            self.__deserialise = value
        else:
            raise TypeError('Expecting callable for "deserialise"; got %r' % 
                            value)
        
    deserialise = property(_getDeserialise, 
                           _setDeserialise, 
                           doc="callable to de-serialise response from XML "
                               "type")        

    def _getIssuer(self):
        return self.__issuer

    def _setIssuer(self, value):
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "issuer"; got %r' %
                            type(value))
            
        self.__issuer = value
        
    issuer = property(fget=_getIssuer, 
                      fset=_setIssuer, 
                      doc="Name of issuing authority")

    def _getIssuerFormat(self):
        if self.__issuerProxy is None:
            return None
        else:
            return self.__issuerProxy.value

    def _setIssuerFormat(self, value):
        if self.__issuerProxy is None:
            self.__issuerProxy = Issuer()
            
        self.__issuerProxy.format = value

    issuerFormat = property(_getIssuerFormat, _setIssuerFormat, 
                            doc="Issuer format")

    def _getIssuerName(self):
        if self.__issuerProxy is None:
            return None
        else:
            return self.__issuerProxy.value

    def _setIssuerName(self, value):
        self.__issuerProxy.value = value

    issuerName = property(_getIssuerName, _setIssuerName, 
                          doc="Name of issuer of SAML Query Response")

    def _getVerifyTimeConditions(self):
        return self.__verifyTimeConditions

    def _setVerifyTimeConditions(self, value):
        if isinstance(value, bool):
            self.__verifyTimeConditions = value
            
        if isinstance(value, basestring):
            self.__verifyTimeConditions = str2Bool(value)
        else:
            raise TypeError('Expecting bool or string type for '
                            '"verifyTimeConditions"; got %r instead' % 
                            type(value))

    verifyTimeConditions = property(_getVerifyTimeConditions, 
                                    _setVerifyTimeConditions, 
                                    doc='Set to True to verify any time '
                                        'Conditions set in the returned '
                                        'response assertions')

    def _getVerifySAMLVersion(self):
        return self.__verifySAMLVersion

    def _setVerifySAMLVersion(self, value):
        if isinstance(value, bool):
            self.__verifySAMLVersion = value
            
        if isinstance(value, basestring):
            self.__verifySAMLVersion = str2Bool(value)
        else:
            raise TypeError('Expecting bool or string type for '
                            '"verifySAMLVersion"; got %r instead' % 
                            type(value))

    verifySAMLVersion = property(_getVerifySAMLVersion, 
                                 _setVerifySAMLVersion, 
                                 doc='Set to True to verify the SAML version '
                                     'set in the query against the SAML '
                                     'Version set in the "samlVersion" '
                                     'attribute')
        
    def _getClockSkewTolerance(self):
        return self.__clockSkewTolerance

    def _setClockSkewTolerance(self, value):
        if isinstance(value, timedelta):
            self.__clockSkewTolerance = value
            
        elif isinstance(value, (float, int, long)):
            self.__clockSkewTolerance = timedelta(seconds=value)
            
        elif isinstance(value, basestring):
            self.__clockSkewTolerance = timedelta(seconds=float(value))
        else:
            raise TypeError('Expecting timedelta, float, int, long or string '
                            'type for "clockSkewTolerance"; got %r' % 
                            type(value))  
                
    clockSkewTolerance = property(fget=_getClockSkewTolerance, 
                                  fset=_setClockSkewTolerance, 
                                  doc="Set a tolerance of +/- n seconds to "
                                      "allow for clock skew when checking the "
                                      "timestamps of client queries")

    def _getSamlVersion(self):
        return self.__samlVersion

    def _setSamlVersion(self, value):
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "samlVersion"; got %r' % 
                            type(value)) 
        self.__samlVersion = value

    samlVersion = property(_getSamlVersion, _setSamlVersion, None, 
                           "SAML Version to enforce for incoming queries.  "
                           "Defaults to version 2.0")
        
    def _getMountPath(self):
        return self.__mountPath
    
    def _setMountPath(self, value):
        '''
        @type value: basestring
        @param value: URL paths to apply this middleware to. Paths are relative 
        to the point at which this middleware is mounted as set in 
        environ['PATH_INFO']
        @raise TypeError: incorrect input type
        '''
        
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "mountPath" attribute; '
                            'got %r' % value)
            
        self.__mountPath = value
            
    mountPath = property(fget=_getMountPath,
                         fset=_setMountPath,
                         doc='URL path to mount this application equivalent to '
                             'environ[\'PATH_INFO\'] (Nb. doesn\'t '
                             'include server domain name or '
                             'environ[\'SCRIPT_NAME\'] setting')
    
    @classmethod
    def filter_app_factory(cls, app, global_conf, **app_conf):
        """Set-up using a Paste app factory pattern.  Set this method to avoid
        possible conflicts from multiple inheritance
        
        @type app: callable following WSGI interface
        @param app: next middleware application in the chain      
        @type global_conf: dict        
        @param global_conf: PasteDeploy global configuration dictionary
        @type prefix: basestring
        @param prefix: prefix for configuration items
        @type app_conf: dict        
        @param app_conf: PasteDeploy application specific configuration 
        dictionary
        """
        app = cls(app)
        app.initialise(global_conf, **app_conf)
        
        return app
    
    def _getQueryInterfaceKeyName(self):
        return self.__queryInterfaceKeyName

    def _setQueryInterfaceKeyName(self, value):
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "queryInterfaceKeyName"'
                            ' got %r' % value)
            
        self.__queryInterfaceKeyName = value

    queryInterfaceKeyName = property(fget=_getQueryInterfaceKeyName, 
                                     fset=_setQueryInterfaceKeyName, 
                                     doc="environ key name for Attribute Query "
                                         "interface")
    
    def __call__(self, environ, start_response):
        """Check for and parse a SOAP SAML Attribute Query and return a
        SAML Response
        
        @type environ: dict
        @param environ: WSGI environment variables dictionary
        @type start_response: function
        @param start_response: standard WSGI start response function
        """
    
        # Ignore non-matching path
        if environ['PATH_INFO'] not in (self.mountPath, 
                                        self.mountPath + '/'):
            return self._app(environ, start_response)
          
        # Ignore non-POST requests
        if environ.get('REQUEST_METHOD') != 'POST':
            return self._app(environ, start_response)
        
        soapRequestStream = environ.get('wsgi.input')
        if soapRequestStream is None:
            raise SOAPQueryInterfaceMiddlewareError('No "wsgi.input" in '
                                                    'environ')
        
        # TODO: allow for chunked data
        contentLength = environ.get('CONTENT_LENGTH')
        if contentLength is None:
            raise SOAPQueryInterfaceMiddlewareError('No "CONTENT_LENGTH" in '
                                                    'environ')

        contentLength = int(contentLength)
        if contentLength <= 0:
            raise SOAPQueryInterfaceMiddlewareError('"CONTENT_LENGTH" in '
                                                    'environ is %d' %
                                                    contentLength)
            
        soapRequestTxt = soapRequestStream.read(contentLength)
        
        # Parse into a SOAP envelope object
        soapRequest = SOAPEnvelope()
        soapRequest.parse(StringIO(soapRequestTxt))
        
        log.debug("SOAPQueryInterfaceMiddleware.__call__: received SAML "
                  "SOAP SQuery ...")
       
        queryElem = soapRequest.body.elem[0]
        
        # Create a response with basic attributes if provided in the 
        # initialisation config
        samlResponse = self._initResponse()
        
        try:
            samlQuery = self.deserialise(queryElem)
            
        except UnknownAttrProfile, e:
            log.exception("%r raised parsing incoming query: %s" % 
                          (type(e), traceback.format_exc()))
            samlResponse.status.statusCode.value = \
                                            StatusCode.UNKNOWN_ATTR_PROFILE_URI
        else:   
            # Check for Query Interface in environ
            queryInterface = environ.get(self.queryInterfaceKeyName,
                                         NotImplemented)
            if queryInterface == NotImplemented:
                raise SOAPQueryInterfaceMiddlewareConfigError(
                                'No query interface %r key found in environ' %
                                self.queryInterfaceKeyName)
                
            elif not callable(queryInterface):
                raise SOAPQueryInterfaceMiddlewareConfigError(
                    'Query interface %r set in %r environ key is not callable' %
                    (queryInterface, self.queryInterfaceKeyName))
            
            # Basic validation
            self._validateQuery(samlQuery, samlResponse)
            
            samlResponse.inResponseTo = samlQuery.id
            
            # Call query interface        
            queryInterface(samlQuery, samlResponse)
        
        # Convert to ElementTree representation to enable attachment to SOAP
        # response body
        samlResponseElem = self.serialise(samlResponse)
        
        # Create SOAP response and attach the SAML Response payload
        soapResponse = SOAPEnvelope()
        soapResponse.create()
        soapResponse.body.elem.append(samlResponseElem)
        
        response = soapResponse.serialize()
        
        log.debug("SOAPQueryInterfaceMiddleware.__call__: sending response "
                  "...\n\n%s",
                  response)
        start_response("200 OK",
                       [('Content-length', str(len(response))),
                        ('Content-type', 'text/xml')])
        return [response]
    
    def _validateQuery(self, query, response):
        """Checking incoming query issue instant and version
        @type query: saml.saml2.core.SubjectQuery 
        @param query: SAML subject query to be checked
        @type: saml.saml2.core.Response
        @param: SAML Response 
        """
        self._verifyQueryTimeConditions(query, response)
        self._verifyQuerySAMLVersion(query, response)
        
    def _verifyQueryTimeConditions(self, query, response):
        """Checking incoming query issue instant
        @type query: saml.saml2.core.SubjectQuery 
        @param query: SAML subject query to be checked
        @type: saml.saml2.core.Response
        @param: SAML Response 
        @raise QueryIssueInstantInvalid: for invalid issue instant
        """
        if not self.verifyTimeConditions: 
            log.debug("Skipping verification of SAML query time conditions")
            return
              
        utcNow = datetime.utcnow() 
        nowPlusSkew = utcNow + self.clockSkewTolerance
        
        if query.issueInstant > nowPlusSkew:
            msg = ('SAML Attribute Query issueInstant [%s] is after '
                   'the clock time [%s] (skewed +%s)' % 
                   (query.issueInstant, 
                    SAMLDateTime.toString(nowPlusSkew),
                    self.clockSkewTolerance))
             
            samlRespError = QueryIssueInstantInvalid(msg)
            samlRespError.response = response
            raise samlRespError
            
    def _verifyQuerySAMLVersion(self, query, response):
        """Checking incoming query issue SAML version
        
        @type query: saml.saml2.core.SubjectQuery 
        @param query: SAML subject query to be checked
        @type: saml.saml2.core.Response
        @param: SAML Response 
        """
        if not self.verifySAMLVersion:
            log.debug("Skipping verification of SAML query version")
            return
        
        if query.version < self.samlVersion:
            msg = ("Query SAML version %r is lower than the supported "
                   "value %r"
                   % (query.version, self.samlVersion))
            response.status.statusCode.value = \
                                        StatusCode.REQUEST_VERSION_TOO_LOW_URI
            return
        
        elif query.version > self.samlVersion:
            msg = ("Query SAML version %r is higher than the supported "
                   "value %r"
                   % (query.version, self.samlVersion))
            response.status.statusCode.value = \
                                        StatusCode.REQUEST_VERSION_TOO_HIGH_URI
            return            
        
    def _initResponse(self):
        """Create a SAML Response object with basic settings if any have been
        provided at initialisation of this class - see initialise
        
        @return: SAML response object
        @rtype: ndg.saml.saml2.core.Response
        """
        samlResponse = Response()
        utcNow = datetime.utcnow()
        
        samlResponse.issueInstant = utcNow
        samlResponse.id = str(uuid4())
        samlResponse.issuer = Issuer()
        
        if self.issuerName is not None:
            samlResponse.issuer.value = self.issuerName
        
        if self.issuerFormat is not None:
            # TODO: Check SAML 2.0 spec says issuer format must be omitted??
            samlResponse.issuer.format = self.issuerFormat
        
        # Initialise to success status but reset on error
        samlResponse.status = Status()
        samlResponse.status.statusCode = StatusCode()
        samlResponse.status.statusMessage = StatusMessage()
        samlResponse.status.statusCode.value = StatusCode.SUCCESS_URI
        
        samlResponse.status.statusMessage = StatusMessage()

        return samlResponse




"""SAML 2.0 bindings module implements SOAP binding for subject query

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "12/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: subjectquery.py 7322 2010-08-13 14:17:22Z pjkersha $'
import logging
log = logging.getLogger(__name__)

from datetime import datetime, timedelta
from uuid import uuid4

from ndg.saml.common import SAMLObject
from ndg.saml.utils import SAMLDateTime
from ndg.saml.saml2.core import (SubjectQuery, StatusCode, Response,
                             Issuer, Subject, SAMLVersion, NameID)

from ndg.saml.utils import str2Bool
from ndg.saml.saml2.binding.soap.client import (SOAPBinding,
    SOAPBindingInvalidResponse)


class SubjectQueryResponseError(SOAPBindingInvalidResponse):
    """SAML Response error from Subject Query"""
    

class IssueInstantInvalid(SubjectQueryResponseError):
    """Issue instant of SAML artifact is invalid"""

  
class ResponseIssueInstantInvalid(IssueInstantInvalid):
    """Issue instant of a response is after the current time"""

    
class AssertionIssueInstantInvalid(IssueInstantInvalid):
    """Issue instant of an assertion is after the current time"""


class AssertionConditionNotBeforeInvalid(SubjectQueryResponseError):
    """An assertion condition notBefore time is set after the current clock
    time"""
    

class AssertionConditionNotOnOrAfterInvalid(SubjectQueryResponseError):
    """An assertion condition notOnOrAfter time is set before the current clock
    time"""

   
class SubjectQuerySOAPBinding(SOAPBinding): 
    """SAML Subject Query SOAP Binding
    """
    SUBJECT_ID_OPTNAME = 'subjectID'
    SUBJECT_ID_FORMAT_OPTNAME = 'subjectIdFormat'
    ISSUER_NAME_OPTNAME = 'issuerName'
    ISSUER_FORMAT_OPTNAME = 'issuerFormat'
    CLOCK_SKEW_OPTNAME = 'clockSkewTolerance'
    VERIFY_TIME_CONDITIONS_OPTNAME = 'verifyTimeConditions'
    
    CONFIG_FILE_OPTNAMES = (
        SUBJECT_ID_OPTNAME,
        SUBJECT_ID_FORMAT_OPTNAME,
        ISSUER_NAME_OPTNAME, 
        ISSUER_FORMAT_OPTNAME,                
        CLOCK_SKEW_OPTNAME,
        VERIFY_TIME_CONDITIONS_OPTNAME            
    )
    
    __PRIVATE_ATTR_PREFIX = "__"
    __slots__ = tuple([__PRIVATE_ATTR_PREFIX + i 
                       for i in CONFIG_FILE_OPTNAMES + ('query', )])
    del i
    
    QUERY_TYPE = SubjectQuery
    
    def __init__(self, **kw):
        '''Create SOAP Client for a SAML Subject Query'''       
        self.__clockSkewTolerance = timedelta(seconds=0.)
        self.__verifyTimeConditions = True
        
        self._initQuery()
        
        super(SubjectQuerySOAPBinding, self).__init__(**kw)

    def _initQuery(self):
        """Initialise query settings"""
        self.__query = self.__class__.QUERY_TYPE()
        self.__query.version = SAMLVersion(SAMLVersion.VERSION_20)
        
        # These properties access the __query instance
        self.issuerFormat = Issuer.X509_SUBJECT
        self.subjectIdFormat = NameID.UNSPECIFIED

    def _getQuery(self):
        return self.__query

    def _setQuery(self, value):
        if not isinstance(value, self.__class__.QUERY_TYPE):
            raise TypeError('Expecting %r query type got %r instead' %
                            (self.__class__, type(value)))
        self.__query = value

    query = property(_getQuery, _setQuery, 
                     doc="SAML Subject Query or derived query type")

    def _getSubjectID(self):
        if self.__query.subject is None or self.__query.subject.nameID is None:
            return None
        else:
            return self.__query.subject.nameID.value

    def _setSubjectID(self, value):
        if self.__query.subject is None:
            self.__query.subject = Subject()
            
        if self.__query.subject.nameID is None:
            self.__query.subject.nameID = NameID()
            
        self.__query.subject.nameID.value = value

    subjectID = property(_getSubjectID, _setSubjectID, 
                         doc="ID to be sent as query subject")
    
    def _getSubjectIdFormat(self):
        if self.__query.subject is None or self.__query.subject.nameID is None:
            return None
        else:
            return self.__query.subject.nameID.format

    def _setSubjectIdFormat(self, value):
        if self.__query.subject is None:
            self.__query.subject = Subject()
            
        if self.__query.subject.nameID is None:
            self.__query.subject.nameID = NameID()
            
        self.__query.subject.nameID.format = value

    subjectIdFormat = property(_getSubjectIdFormat, _setSubjectIdFormat, 
                               doc="Subject Name ID format")

    def _getIssuerFormat(self):
        if self.__query.issuer is None:
            return None
        else:
            return self.__query.issuer.value

    def _setIssuerFormat(self, value):
        if self.__query.issuer is None:
            self.__query.issuer = Issuer()
            
        self.__query.issuer.format = value

    issuerFormat = property(_getIssuerFormat, _setIssuerFormat, 
                            doc="Issuer format")

    def _getIssuerName(self):
        if self.__query.issuer is None:
            return None
        else:
            return self.__query.issuer.value

    def _setIssuerName(self, value):
        if self.__query.issuer is None:
            self.__query.issuer = Issuer()
            
        self.__query.issuer.value = value

    issuerName = property(_getIssuerName, _setIssuerName, 
                          doc="Name of issuer of SAML Subject Query")

    def _getVerifyTimeConditions(self):
        return self.__verifyTimeConditions

    def _setVerifyTimeConditions(self, value):
        if isinstance(value, bool):
            self.__verifyTimeConditions = value
            
        if isinstance(value, basestring):
            self.__verifyTimeConditions = str2Bool(value)
        else:
            raise TypeError('Expecting bool or string type for '
                            '"verifyTimeConditions"; got %r instead' % 
                            type(value))

    verifyTimeConditions = property(_getVerifyTimeConditions, 
                                    _setVerifyTimeConditions, 
                                    doc='Set to True to verify any time '
                                        'Conditions set in the returned '
                                        'response assertions')  

    def _getClockSkewTolerance(self):
        return self.__clockSkewTolerance

    def _setClockSkewTolerance(self, value):
        if isinstance(value, timedelta):
            self.__clockSkewTolerance = value
            
        elif isinstance(value, (float, int, long)):
            self.__clockSkewTolerance = timedelta(seconds=value)
            
        elif isinstance(value, basestring):
            self.__clockSkewTolerance = timedelta(seconds=float(value))
        else:
            raise TypeError('Expecting timedelta, float, int, long or string '
                            'type for "clockSkewTolerance"; got %r' % 
                            type(value))

    clockSkewTolerance = property(fget=_getClockSkewTolerance, 
                                  fset=_setClockSkewTolerance, 
                                  doc="Allow a tolerance in seconds for SAML "
                                      "Query issueInstant parameter check and "
                                      "assertion condition notBefore and "
                                      "notOnOrAfter times to allow for clock "
                                      "skew")
    
    def _validateQueryParameters(self):
        """Perform sanity check immediately before creating the query and 
        sending it"""
        errors = []
        
        if self.issuerName is None:
            errors.append('issuer name')

        if self.issuerFormat is None:
            errors.append('issuer format')
        
        if self.subjectID is None:
            errors.append('subject')
        
        if self.subjectIdFormat is None:
            errors.append('subject format')
        
        if errors:
            raise AttributeError('Missing attribute(s) for SAML Query: %s' %
                                 ', '.join(errors))

    def _initSend(self):
        """Perform any final initialisation prior to sending the query - derived
        classes may overload to specify as required"""
        self.__query.issueInstant = datetime.utcnow()
        
        # Set ID here to ensure it's unique for each new call
        self.__query.id = str(uuid4())

    def _verifyTimeConditions(self, response):
        """Verify time conditions set in a response
        @param response: SAML Response returned from remote service
        @type response: ndg.saml.saml2.core.Response
        @raise SubjectQueryResponseError: if a timestamp is invalid
        """
        
        if not self.verifyTimeConditions:
            log.debug("Skipping verification of SAML Response time conditions")
            
        utcNow = datetime.utcnow() 
        nowMinusSkew = utcNow - self.clockSkewTolerance
        nowPlusSkew = utcNow + self.clockSkewTolerance
        
        if response.issueInstant > nowPlusSkew:
            msg = ('SAML Attribute Response issueInstant [%s] is after '
                   'the clock time [%s] (skewed +%s)' % 
                   (response.issueInstant, 
                    SAMLDateTime.toString(nowPlusSkew),
                    self.clockSkewTolerance))
             
            samlRespError = ResponseIssueInstantInvalid(msg)
            samlRespError.response = response
            raise samlRespError
        
        for assertion in response.assertions:
            if assertion.issueInstant is None:
                samlRespError = AssertionIssueInstantInvalid("No issueInstant "
                                                             "set in response "
                                                             "assertion")
                samlRespError.response = response
                raise samlRespError
            
            elif nowPlusSkew < assertion.issueInstant:
                msg = ('The clock time [%s] (skewed +%s) is before the '
                       'SAML Attribute Response assertion issue instant [%s]' % 
                       (SAMLDateTime.toString(utcNow),
                        self.clockSkewTolerance,
                        assertion.issueInstant))
                samlRespError = AssertionIssueInstantInvalid(msg)
                samlRespError.response = response
                raise samlRespError
            
            if assertion.conditions is not None:
                if nowPlusSkew < assertion.conditions.notBefore:            
                    msg = ('The clock time [%s] (skewed +%s) is before the '
                           'SAML Attribute Response assertion conditions not '
                           'before time [%s]' % 
                           (SAMLDateTime.toString(utcNow),
                            self.clockSkewTolerance,
                            assertion.conditions.notBefore))
                              
                    samlRespError = AssertionConditionNotBeforeInvalid(msg)
                    samlRespError.response = response
                    raise samlRespError
                 
                if nowMinusSkew >= assertion.conditions.notOnOrAfter:           
                    msg = ('The clock time [%s] (skewed -%s) is on or after '
                           'the SAML Attribute Response assertion conditions '
                           'not on or after time [%s]' % 
                           (SAMLDateTime.toString(utcNow),
                            self.clockSkewTolerance,
                            assertion.conditions.notOnOrAfter))
                    
                    samlRespError = AssertionConditionNotOnOrAfterInvalid(msg) 
                    samlRespError.response = response
                    raise samlRespError
                
    def send(self, **kw):
        '''Make an attribute query to a remote SAML service
        
        @type uri: basestring 
        @param uri: uri of service.  May be omitted if set from request.url
        @type request: ndg.security.common.soap.UrlLib2SOAPRequest
        @param request: SOAP request object to which query will be attached
        defaults to ndg.security.common.soap.client.UrlLib2SOAPRequest
        '''
        self._validateQueryParameters() 
        self._initSend()
           
        response = super(SubjectQuerySOAPBinding, self).send(self.query, **kw)

        # Perform validation
        if response.status.statusCode.value != StatusCode.SUCCESS_URI:
            msg = ('Return status code flagged an error, %r.  '
                   'The message is, %r' %
                   (response.status.statusCode.value,
                    response.status.statusMessage.value))
            samlRespError = SubjectQueryResponseError(msg)
            samlRespError.response = response
            raise samlRespError
        
        # Check Query ID matches the query ID the service received
        if response.inResponseTo != self.query.id:
            msg = ('Response in-response-to ID %r, doesn\'t match the original '
                   'query ID, %r' % (response.inResponseTo, query.id))
            
            samlRespError = SubjectQueryResponseError(msg)
            samlRespError.response = response
            raise samlRespError
        
        self._verifyTimeConditions(response)
            
        return response 



"""ElementTree Utilities package for NDG SOAP Package

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "02/04/09"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: etree.py 7131 2010-06-30 13:37:48Z pjkersha $'
try: # python 2.5
    from xml.etree import ElementTree
except ImportError:
    # if you've installed it yourself it comes this way
    import ElementTree

import re


class QName(ElementTree.QName):
    """XML Qualified Name for ElementTree
    
    Extends ElementTree implementation for improved attribute access support
    """ 
    # ElementTree tag is of the form {namespace}localPart.  getNs extracts the
    # namespace from within the brackets but if not found returns ''
    getNs = staticmethod(lambda tag: getattr(re.search('(?<=\{).+(?=\})', tag),
                                             'group', 
                                             str)())
                                             
    getLocalPart = staticmethod(lambda tag: tag.rsplit('}', 1)[-1])
    
    def __init__(self, namespaceURI, tag=None, prefix=None):
        """Initialise a qualified name
        
        @param namespaceURI: element namespace URI
        @type namespaceURI: basestring
        @param tag: element local name
        @type tag: basestring
        @param prefix: XML namespace prefix
        @type prefix: basestring
        """
        ElementTree.QName.__init__(self, namespaceURI, tag=tag)
        
        if tag:
            self.namespaceURI = namespaceURI
            self.localPart = tag
        else:
            self.namespaceURI = QName.getNs(namespaceURI)
            self.localPart = QName.getLocalPart(namespaceURI)
            
        self.prefix = prefix

    def __eq__(self, qname):
        """Enable equality check for QName
        @type qname: ndg.security.common.utils.etree.QName
        @param qname: Qualified Name to compare with self 
        @return: True if names are equal
        @rtype: bool
        """
        if not isinstance(qname, QName):
            raise TypeError('Expecting %r; got %r' % (QName, type(qname)))
                            
        return (self.prefix, self.namespaceURI, self.localPart) == \
               (qname.prefix, qname.namespaceURI, qname.localPart)

    def __ne__(self, qname):
        """Enable equality check for QName
        @type qname: ndg.security.common.utils.etree.QName
        @param qname: Qualified Name to compare with self 
        @return: True if names are not equal
        @rtype: bool
        """
        return not self.__eq__(qname)
               
    def _getPrefix(self):
        return self.__prefix

    def _setPrefix(self, value):
        self.__prefix = value
    
    prefix = property(_getPrefix, _setPrefix, None, "Prefix")

    def _getLocalPart(self):
        return self.__localPart
    
    def _setLocalPart(self, value):
        self.__localPart = value
        
    localPart = property(_getLocalPart, _setLocalPart, None, "LocalPart")

    def _getNamespaceURI(self):
        return self.__namespaceURI

    def _setNamespaceURI(self, value):
        self.__namespaceURI = value
  
    namespaceURI = property(_getNamespaceURI, _setNamespaceURI, None, 
                            "Namespace URI'")


def prettyPrint(*arg, **kw):
    '''Lightweight pretty printing of ElementTree elements'''
    
    # Keep track of namespace declarations made so they're not repeated
    declaredNss = []
    
    _prettyPrint = _PrettyPrint(declaredNss)
    return _prettyPrint(*arg, **kw)


class _PrettyPrint(object):
    def __init__(self, declaredNss):
        self.declaredNss = declaredNss
    
    @staticmethod
    def estrip(elem):
        ''' Just want to get rid of unwanted whitespace '''
        if elem is None:
            return ''
        else:
            # just in case the elem is another simple type - e.g. int - 
            # wrapper it as a string
            return str(elem).strip()
        
    def __call__(self, elem, indent='', html=0, space=' '*4):
        '''Most of the work done in this wrapped function - wrapped so that
        state can be maintained for declared namespace declarations during
        recursive calls using "declaredNss" above'''  
        strAttribs = []
        for attr, attrVal in elem.attrib.items():
            nsDeclaration = ''
            
            attrNamespace = QName.getNs(attr)
            if attrNamespace:
                nsPrefix = ElementTree._namespace_map.get(attrNamespace)
                if nsPrefix is None:
                    raise KeyError('prettyPrint: missing namespace "%s" for ' 
                                   'ElementTree._namespace_map'%attrNamespace)
                
                attr = "%s:%s" % (nsPrefix, QName.getLocalPart(attr))
                
                if attrNamespace not in self.declaredNss:
                    nsDeclaration = ' xmlns:%s="%s"' % (nsPrefix,attrNamespace)
                    self.declaredNss.append(attrNamespace)
                
            strAttribs.append('%s %s="%s"' % (nsDeclaration, attr, attrVal))
            
        strAttrib = ''.join(strAttribs)
        
        namespace = QName.getNs(elem.tag)
        nsPrefix = ElementTree._namespace_map.get(namespace)
        if nsPrefix is None:
            raise KeyError('prettyPrint: missing namespace "%s" for ' 
                           'ElementTree._namespace_map' % namespace)
            
        tag = "%s:%s" % (nsPrefix, QName.getLocalPart(elem.tag))
        
        # Put in namespace declaration if one doesn't already exist
        # FIXME: namespace declaration handling is wrong for handling child
        # element scope
        if namespace in self.declaredNss:
            nsDeclaration = ''
        else:
            nsDeclaration = ' xmlns:%s="%s"' % (nsPrefix, namespace)
            self.declaredNss.append(namespace)
            
        result = '%s<%s%s%s>%s' % (indent, tag, nsDeclaration, strAttrib, 
                                   _PrettyPrint.estrip(elem.text))
        
        children = len(elem)
        if children:
            for child in elem:
                declaredNss = self.declaredNss[:]
                _prettyPrint = _PrettyPrint(declaredNss)
                result += '\n'+ _prettyPrint(child, indent=indent+space) 
                
            result += '\n%s%s</%s>' % (indent,
                                     _PrettyPrint.estrip(child.tail),
                                     tag)
        else:
            result += '</%s>' % tag
            
        return result




"""NDG SAML 2.0 ndg namespace package

NERC DataGrid Project

This is a setuptools namespace_package.  DO NOT place any other
code in this file!  There is no guarantee that it will be installed
with easy_install.  See:

http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages

... for details.
"""
__author__ = "P J Kershaw"
__date__ = "23/02/2010"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $'

__import__('pkg_resources').declare_namespace(__name__)


"""NDG SAML SOAP Binding unit test package

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "21/08/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7154 2010-07-01 15:59:59Z pjkersha $'
import os
import unittest
import socket
import paste.fixture
from paste.deploy import loadapp

from ndg.soap.test import PasteDeployAppServer


class TestApp(object):
    """Dummy application to terminate middleware stack containing SAML service
    """
    def __init__(self, global_conf, **app_conf):
        pass
    
    def __call__(self, environ, start_response):
        response = "404 Not Found"
        start_response(response,
                       [('Content-length', str(len(response))),
                        ('Content-type', 'text/plain')])
                            
        return [response]


class WithPasteFixtureBaseTestCase(unittest.TestCase):
    """Base class for testing SAML SOAP Binding Query/Response interface
    using a Paste Deploy ini file and Paste Fixture
    """
    HERE_DIR = os.path.dirname(os.path.abspath(__file__))
    CONFIG_FILENAME = None # Set in derived class
    
    def __init__(self, *args, **kwargs):
        wsgiapp = loadapp('config:%s' % self.__class__.CONFIG_FILENAME, 
                          relative_to=self.__class__.HERE_DIR)
        
        self.app = paste.fixture.TestApp(wsgiapp)
         
        unittest.TestCase.__init__(self, *args, **kwargs)
        
    
class WithPasterBaseTestCase(unittest.TestCase):
    """Base class for testing SAML SOAP Binding Query/Response interface
    using a Paste Deploy ini file and Paste Fixture
    """
    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    CONFIG_FILENAME = None # Set in derived class
    SERVICE_PORTNUM = 5443
    SERVER_CERT_FILEPATH = os.path.join(THIS_DIR, 'localhost.crt')
    SERVER_PRIKEY_FILEPATH = os.path.join(THIS_DIR, 'localhost.key')
    
    def __init__(self, *arg, **kw):
        withSSL = kw.pop('withSSL', False)
        unittest.TestCase.__init__(self, *arg, **kw)
        
        self.services = []
        self.disableServiceStartup = False
        cfgFilePath = os.path.join(self.__class__.THIS_DIR, 
                                   self.__class__.CONFIG_FILENAME)
        
        self.addService(cfgFilePath=cfgFilePath,
                        withSSL=withSSL,
                        port=self.__class__.SERVICE_PORTNUM)
        
    def addService(self, *arg, **kw):
        """Utility for setting up threads to run Paste HTTP based services with
        unit tests
        
        @param arg: tuple contains ini file path setting for the service
        @type arg: tuple
        @param kw: keywords including "port" - port number to run the service 
        from
        @type kw: dict
        """
        if self.disableServiceStartup:
            return
        
        withSSL = kw.pop('withSSL', False)
        if withSSL:
            from OpenSSL import SSL
            
            certFilePath = self.__class__.SERVER_CERT_FILEPATH
            priKeyFilePath = self.__class__.SERVER_PRIKEY_FILEPATH
            
            kw['ssl_context'] = SSL.Context(SSL.SSLv23_METHOD)
            kw['ssl_context'].set_options(SSL.OP_NO_SSLv2)
        
            kw['ssl_context'].use_privatekey_file(priKeyFilePath)
            kw['ssl_context'].use_certificate_file(certFilePath)
        
        try:
            self.services.append(PasteDeployAppServer(*arg, **kw))
            self.services[-1].startThread()
            
        except socket.error:
            pass

    def __del__(self):
        """Stop any services started with the addService method and clean up
        the CA directory following the trust roots call
        """
        if hasattr(self, 'services'):
            for service in self.services:
                service.terminateThread()



"""SAML 2.0 SOAP Binding package

Implementation of SAML 2.0 for NDG Security

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "30/06/10"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7259 2010-08-02 12:26:32Z pjkersha $"
from ndg.saml.saml2.core import Response

                                 
class SOAPBindingError(Exception):
    '''Base exception type for client SAML SOAP Binding'''


class SOAPBindingInvalidResponse(SOAPBindingError):
    '''Raise if the response is invalid'''
    def __init__(self, *arg, **kw):
        SOAPBindingError.__init__(self, *arg, **kw)
        self.__response = None
    
    def _getResponse(self):
        '''Gets the response corresponding to this error
        
        @return the response
        '''
        return self.__response

    def _setResponse(self, value):
        '''Sets the response corresponding to this error.
        
        @param value: the response
        '''
        if not isinstance(value, Response):
            raise TypeError('"response" must be a %r, got %r' % (Response,
                                                                 type(value)))
        self.__response = value
        
    response = property(fget=_getResponse, fset=_setResponse, 
                        doc="SAML Response associated with this exception")



"""SOAP common package for NDG SAML.  

Initially for use with SAML SOAP Bindings.  This itself
uses ElementTree.  This SOAP interface provides an ElementTree interface to
support it

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "24/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $'
import logging
log = logging.getLogger(__name__)

class SOAPException(Exception):
    """Base SAOP Exception class"""
    
class SOAPFault(SOAPException):
    """SOAP Fault"""
    
class SOAPObject(object):
    """Base class for SOAP envelope, header and body elements"""
    
    ELEMENT_PREFIX = "soap11"
    SOAP11_NS = "http://schemas.xmlsoap.org/soap/envelope/"
    SOAP12_NS = "http://www.w3.org/2003/05/soap-envelope"
    DEFAULT_NS = SOAP11_NS
    
    def create(self):
        raise NotImplementedError()
    
    def parse(self):
        raise NotImplementedError()
    
    def serialize(self):
        raise NotImplementedError()
    
    def prettyPrint(self):
        raise NotImplementedError()
  
    
class SOAPEnvelopeBase(SOAPObject):
    """SOAP Envelope"""
    
    DEFAULT_ELEMENT_LOCAL_NAME = "Envelope"
    DEFAULT_ELEMENT_NS = SOAPObject.DEFAULT_NS
    DEFAULT_ELEMENT_NS_PREFIX = SOAPObject.ELEMENT_PREFIX
    
    soapHeader = property()
    soapBody = property()
    
    
class SOAPHeaderBase(SOAPObject):
    """SOAP Header base class"""
    
    DEFAULT_ELEMENT_LOCAL_NAME = "Header"
    DEFAULT_ELEMENT_NS = SOAPObject.DEFAULT_NS
    DEFAULT_ELEMENT_NS_PREFIX = SOAPObject.ELEMENT_PREFIX    
        
class SOAPBodyBase(SOAPObject):
    """SOAP Body base class"""
    
    DEFAULT_ELEMENT_LOCAL_NAME = "Body"
    DEFAULT_ELEMENT_NS = SOAPObject.DEFAULT_NS
    DEFAULT_ELEMENT_NS_PREFIX = SOAPObject.ELEMENT_PREFIX



"""SAML unit test package

NERC DataGrid Project

This implementation is adapted from the Java OpenSAML implementation.  The 
copyright and licence information are included here:

Copyright [2005] [University Corporation for Advanced Internet Development, Inc.]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__author__ = "P J Kershaw"
__date__ = "21/07/09"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: test_saml.py 7130 2010-06-30 13:33:07Z pjkersha $'
import logging
logging.basicConfig(level=logging.DEBUG)
    
from datetime import datetime, timedelta
from uuid import uuid4
from cStringIO import StringIO

import unittest
import pickle

from xml.etree.ElementTree import iselement
from xml.etree import ElementTree

from ndg.saml.saml2.core import (SAMLVersion, Attribute, AttributeStatement, 
                                 AuthzDecisionStatement, Assertion, 
                                 AttributeQuery, Response, Issuer, Subject, 
                                 NameID, StatusCode, StatusMessage, Status, 
                                 Conditions, DecisionType, 
                                 XSStringAttributeValue, Action, 
                                 AuthzDecisionQuery)

from ndg.saml.common.xml import SAMLConstants
from ndg.saml.xml.etree import (prettyPrint, AssertionElementTree, 
                            AttributeQueryElementTree, ResponseElementTree,
                            AuthzDecisionQueryElementTree)


class SAMLUtil(object):
    """SAML utility class based on ANL examples for Earth System Grid:
    http://www.ci.uchicago.edu/wiki/bin/view/ESGProject/ESGSAMLAttributes#ESG_Attribute_Service
    """
    NAMEID_FORMAT = "urn:esg:openid"
    NAMEID_VALUE = "https://openid.localhost/philip.kershaw"
    ISSUER_DN = "/O=NDG/OU=BADC/CN=attributeauthority.badc.rl.ac.uk"
    UNCORRECTED_RESOURCE_URI = "http://LOCALHOST:80/My Secured URI"
    RESOURCE_URI = "http://localhost/My%20Secured%20URI"
    XSSTRING_NS = "http://www.w3.org/2001/XMLSchema#string"
    
    def __init__(self):
        """Set-up ESG core attributes, Group/Role and miscellaneous 
        attributes lists
        """
        self.firstName = None
        self.lastName = None
        self.emailAddress = None
        
        self.__miscAttrList = []
    
    def addAttribute(self, name, value):
        """Add a generic attribute
        @type name: basestring
        @param name: attribute name
        @type value: basestring
        @param value: attribute value
        """
        self.__miscAttrList.append((name, value))

    def buildAssertion(self):
        """Create a SAML Assertion containing ESG core attributes: First
        Name, Last Name, e-mail Address; ESG Group/Role type attributes
        and generic attributes
        @rtype: ndg.security.common.saml.Assertion
        @return: new SAML Assertion object
        """
        
        assertion = Assertion()
        assertion.version = SAMLVersion(SAMLVersion.VERSION_20)
        assertion.id = str(uuid4())
        assertion.issueInstant = datetime.utcnow()
        attributeStatement = AttributeStatement()
        
        for attribute in self.createAttributes():
            attributeStatement.attributes.append(attribute)
            
        assertion.attributeStatements.append(attributeStatement)
        
        return assertion

    def buildAttributeQuery(self, issuer, subjectNameID):
        """Make a SAML Attribute Query
        @type issuer: basestring
        @param issuer: attribute issuer name
        @type subjectNameID: basestring
        @param subjectNameID: identity to query attributes for
        """
        attributeQuery = AttributeQuery()
        attributeQuery.version = SAMLVersion(SAMLVersion.VERSION_20)
        attributeQuery.id = str(uuid4())
        attributeQuery.issueInstant = datetime.utcnow()
        
        attributeQuery.issuer = Issuer()
        attributeQuery.issuer.format = Issuer.X509_SUBJECT
        attributeQuery.issuer.value = issuer
                        
        attributeQuery.subject = Subject()  
        attributeQuery.subject.nameID = NameID()
        attributeQuery.subject.nameID.format = SAMLTestCase.NAMEID_FORMAT
        attributeQuery.subject.nameID.value = subjectNameID
                                    
        attributeQuery.attributes = self.createAttributes()
        
        return attributeQuery
    
    def createAttributes(self):
        """Create SAML Attributes for use in an Assertion or AttributeQuery"""
        
        attributes = []
        if self.firstName is not None:    
            # special case handling for 'FirstName' attribute
            fnAttribute = Attribute()
            fnAttribute.name = "urn:esg:first:name"
            fnAttribute.nameFormat = SAMLUtil.XSSTRING_NS
            fnAttribute.friendlyName = "FirstName"

            firstName = XSStringAttributeValue()
            firstName.value = self.firstName
            fnAttribute.attributeValues.append(firstName)

            attributes.append(fnAttribute)
        

        if self.lastName is not None:
            # special case handling for 'LastName' attribute
            lnAttribute = Attribute()
            lnAttribute.name = "urn:esg:last:name"
            lnAttribute.nameFormat = SAMLUtil.XSSTRING_NS
            lnAttribute.friendlyName = "LastName"

            lastName = XSStringAttributeValue()
            lastName.value = self.lastName
            lnAttribute.attributeValues.append(lastName)

            attributes.append(lnAttribute)
        

        if self.emailAddress is not None:
            # special case handling for 'LastName' attribute
            emailAddressAttribute = Attribute()
            emailAddressAttribute.name = "urn:esg:email:address"
            emailAddressAttribute.nameFormat = SAMLConstants.XSD_NS+"#"+\
                                        XSStringAttributeValue.TYPE_LOCAL_NAME
            emailAddressAttribute.friendlyName = "emailAddress"

            emailAddress = XSStringAttributeValue()
            emailAddress.value = self.emailAddress
            emailAddressAttribute.attributeValues.append(emailAddress)

            attributes.append(emailAddressAttribute)
        
        for name, value in self.__miscAttrList:
            attribute = Attribute()
            attribute.name = name
            attribute.nameFormat = SAMLUtil.XSSTRING_NS

            stringAttributeValue = XSStringAttributeValue()
            stringAttributeValue.value = value
            attribute.attributeValues.append(stringAttributeValue)

            attributes.append(attribute)
            
        return attributes
    
    def buildAuthzDecisionQuery(self, 
                                issuer=ISSUER_DN,
                                issuerFormat=Issuer.X509_SUBJECT,
                                subjectNameID=NAMEID_VALUE, 
                                subjectNameIdFormat=NAMEID_FORMAT,
                                resource=UNCORRECTED_RESOURCE_URI,
                                actions=((Action.HTTP_GET_ACTION, 
                                          Action.GHPP_NS_URI),)):
        """Convenience utility to make an Authorisation decision query"""
        authzDecisionQuery = AuthzDecisionQuery()

        authzDecisionQuery.version = SAMLVersion(SAMLVersion.VERSION_20)
        authzDecisionQuery.id = str(uuid4())
        authzDecisionQuery.issueInstant = datetime.utcnow()
        
        authzDecisionQuery.issuer = Issuer()
        authzDecisionQuery.issuer.format = issuerFormat
        authzDecisionQuery.issuer.value = issuer
        
        authzDecisionQuery.subject = Subject()
        authzDecisionQuery.subject.nameID = NameID()
        authzDecisionQuery.subject.nameID.format = subjectNameIdFormat
        authzDecisionQuery.subject.nameID.value = subjectNameID
        
        authzDecisionQuery.resource = resource
        
        for action, namespace in actions:
            authzDecisionQuery.actions.append(Action())
            authzDecisionQuery.actions[-1].namespace = namespace
            authzDecisionQuery.actions[-1].value = action
            
        return authzDecisionQuery
            

class SAMLTestCase(unittest.TestCase):
    """Test SAML implementation for use with CMIP5 federation"""
    NAMEID_FORMAT = SAMLUtil.NAMEID_FORMAT
    NAMEID_VALUE = SAMLUtil.NAMEID_VALUE
    ISSUER_DN = SAMLUtil.ISSUER_DN
    UNCORRECTED_RESOURCE_URI = SAMLUtil.UNCORRECTED_RESOURCE_URI
    RESOURCE_URI = SAMLUtil.RESOURCE_URI
    
    def _createAttributeAssertionHelper(self):
        samlUtil = SAMLUtil()
        
        # ESG core attributes
        samlUtil.firstName = "Philip"
        samlUtil.lastName = "Kershaw"
        samlUtil.emailAddress = "p.j.k@somewhere"
        
        # BADC specific attributes
        badcRoleList = (
            'urn:badc:security:authz:1.0:attr:admin', 
            'urn:badc:security:authz:1.0:attr:rapid', 
            'urn:badc:security:authz:1.0:attr:coapec', 
            'urn:badc:security:authz:1.0:attr:midas', 
            'urn:badc:security:authz:1.0:attr:quest', 
            'urn:badc:security:authz:1.0:attr:staff'
        )
        for role in badcRoleList:
            samlUtil.addAttribute("urn:badc:security:authz:1.0:attr", role)
        
        # Make an assertion object
        assertion = samlUtil.buildAssertion()
        
        return assertion
        
    def test01CreateAssertion(self):
         
        assertion = self._createAttributeAssertionHelper()

        
        # Create ElementTree Assertion Element
        assertionElem = AssertionElementTree.toXML(assertion)
        
        self.assert_(iselement(assertionElem))
        
        # Serialise to output 
        xmlOutput = prettyPrint(assertionElem)       
        self.assert_(len(xmlOutput))
        
        print("\n"+"_"*80)
        print(xmlOutput)
        print("_"*80)

    def test02ParseAssertion(self):
        assertion = self._createAttributeAssertionHelper()
        
        # Create ElementTree Assertion Element
        assertionElem = AssertionElementTree.toXML(assertion)
        
        self.assert_(iselement(assertionElem))
        
        # Serialise to output 
        xmlOutput = prettyPrint(assertionElem)       
           
        print("\n"+"_"*80)
        print(xmlOutput)
        print("_"*80)
                
        assertionStream = StringIO()
        assertionStream.write(xmlOutput)
        assertionStream.seek(0)

        tree = ElementTree.parse(assertionStream)
        elem2 = tree.getroot()
        
        assertion2 = AssertionElementTree.fromXML(elem2)
        self.assert_(assertion2)
        
    def test03CreateAttributeQuery(self):
        samlUtil = SAMLUtil()
        samlUtil.firstName = ''
        samlUtil.lastName = ''
        samlUtil.emailAddress = ''
        attributeQuery = samlUtil.buildAttributeQuery(SAMLTestCase.ISSUER_DN,
                                                      SAMLTestCase.NAMEID_VALUE)
        
        elem = AttributeQueryElementTree.toXML(attributeQuery)
        xmlOutput = prettyPrint(elem)
           
        print("\n"+"_"*80)
        print(xmlOutput)
        print("_"*80)

    def test04ParseAttributeQuery(self):
        samlUtil = SAMLUtil()
        samlUtil.firstName = ''
        samlUtil.lastName = ''
        samlUtil.emailAddress = ''
        attributeQuery = samlUtil.buildAttributeQuery(SAMLTestCase.ISSUER_DN,
                                                      SAMLTestCase.NAMEID_VALUE)
        
        elem = AttributeQueryElementTree.toXML(attributeQuery)        
        xmlOutput = prettyPrint(elem)       
        print("\n"+"_"*80)
        print(xmlOutput)
                
        attributeQueryStream = StringIO()
        attributeQueryStream.write(xmlOutput)
        attributeQueryStream.seek(0)

        tree = ElementTree.parse(attributeQueryStream)
        elem2 = tree.getroot()
        
        attributeQuery2 = AttributeQueryElementTree.fromXML(elem2)
        self.assert_(attributeQuery2.id == attributeQuery.id)
        self.assert_(attributeQuery2.issuer.value==attributeQuery.issuer.value)
        self.assert_(attributeQuery2.subject.nameID.value == \
                     attributeQuery.subject.nameID.value)
        
        self.assert_(attributeQuery2.attributes[1].name == \
                     attributeQuery.attributes[1].name)
        
        xmlOutput2 = prettyPrint(elem2)       
        print("_"*80)
        print(xmlOutput2)
        print("_"*80)

    def _createAttributeQueryResponse(self):
        response = Response()
        response.issueInstant = datetime.utcnow()
        
        # Make up a request ID that this response is responding to
        response.inResponseTo = str(uuid4())
        response.id = str(uuid4())
        response.version = SAMLVersion(SAMLVersion.VERSION_20)
            
        response.issuer = Issuer()
        response.issuer.format = Issuer.X509_SUBJECT
        response.issuer.value = \
                        SAMLTestCase.ISSUER_DN
        
        response.status = Status()
        response.status.statusCode = StatusCode()
        response.status.statusCode.value = StatusCode.SUCCESS_URI
        response.status.statusMessage = StatusMessage()        
        response.status.statusMessage.value = "Response created successfully"
           
        assertion = self._createAttributeAssertionHelper()
        
        # Add a conditions statement for a validity of 8 hours
        assertion.conditions = Conditions()
        assertion.conditions.notBefore = datetime.utcnow()
        assertion.conditions.notOnOrAfter = (assertion.conditions.notBefore + 
                                             timedelta(seconds=60*60*8))
        
        assertion.subject = Subject()  
        assertion.subject.nameID = NameID()
        assertion.subject.nameID.format = SAMLTestCase.NAMEID_FORMAT
        assertion.subject.nameID.value = SAMLTestCase.NAMEID_VALUE    
            
        assertion.issuer = Issuer()
        assertion.issuer.format = Issuer.X509_SUBJECT
        assertion.issuer.value = SAMLTestCase.ISSUER_DN

        response.assertions.append(assertion)
        
        return response
        
    def test05CreateAttributeQueryResponse(self):
        response = self._createAttributeQueryResponse()
        
        # Create ElementTree Assertion Element
        responseElem = ResponseElementTree.toXML(response)
        
        self.assert_(iselement(responseElem))
        
        # Serialise to output        
        xmlOutput = prettyPrint(responseElem)       
        self.assert_(len(xmlOutput))
        print("\n"+"_"*80)
        print(xmlOutput)
        print("_"*80)
        
    def _createAuthzDecisionQuery(self):
        authzDecisionQuery = AuthzDecisionQuery()

        authzDecisionQuery.version = SAMLVersion(SAMLVersion.VERSION_20)
        authzDecisionQuery.id = str(uuid4())
        authzDecisionQuery.issueInstant = datetime.utcnow()
        
        authzDecisionQuery.issuer = Issuer()
        authzDecisionQuery.issuer.format = Issuer.X509_SUBJECT
        authzDecisionQuery.issuer.value = SAMLTestCase.ISSUER_DN
        
        authzDecisionQuery.subject = Subject()
        authzDecisionQuery.subject.nameID = NameID()
        authzDecisionQuery.subject.nameID.format = SAMLTestCase.NAMEID_FORMAT
        authzDecisionQuery.subject.nameID.value = SAMLTestCase.NAMEID_VALUE
        
        authzDecisionQuery.resource = "http://LOCALHOST:80/My Secured URI"
        
        return authzDecisionQuery
    
    def test06CreateAuthzDecisionQuery(self):
        samlUtil = SAMLUtil()
        authzDecisionQuery = samlUtil.buildAuthzDecisionQuery()
        
        self.assert_(":80" not in authzDecisionQuery.resource)
        self.assert_("localhost" in authzDecisionQuery.resource)
        self.assert_(" " not in authzDecisionQuery.resource)
        
        authzDecisionQuery.resource = \
            "https://Somewhere.ac.uk:443/My Secured URI?blah=4&yes=True"
            
        self.assert_(":443" not in authzDecisionQuery.resource)
        self.assert_("somewhere.ac.uk" in authzDecisionQuery.resource)
        self.assert_("yes=True" in authzDecisionQuery.resource)
        
        authzDecisionQuery.actions.append(Action())
        authzDecisionQuery.actions[0].namespace = Action.GHPP_NS_URI
        authzDecisionQuery.actions[0].value = Action.HTTP_GET_ACTION
        
        self.assert_(
            authzDecisionQuery.actions[0].value == Action.HTTP_GET_ACTION)
        self.assert_(
            authzDecisionQuery.actions[0].namespace == Action.GHPP_NS_URI)
        
        # Try out the restricted vocabulary
        try:
            authzDecisionQuery.actions[0].value = "delete everything"
            self.fail("Expecting AttributeError raised for incorrect action "
                      "setting.")
        except AttributeError, e:
            print("Caught incorrect action type setting: %s" % e)
        
        authzDecisionQuery.actions[0].actionTypes = {'urn:malicious': 
                                                     ("delete everything",)}
        
        # Try again now that the actipn types have been adjusted
        authzDecisionQuery.actions[0].namespace = 'urn:malicious'
        authzDecisionQuery.actions[0].value = "delete everything"
        
    def test07SerializeAuthzDecisionQuery(self):
        samlUtil = SAMLUtil()
        authzDecisionQuery = samlUtil.buildAuthzDecisionQuery()
        
        # Create ElementTree Assertion Element
        authzDecisionQueryElem = AuthzDecisionQueryElementTree.toXML(
                                                            authzDecisionQuery)
        
        self.assert_(iselement(authzDecisionQueryElem))
        
        # Serialise to output 
        xmlOutput = prettyPrint(authzDecisionQueryElem)       
        self.assert_(len(xmlOutput))
        
        print("\n"+"_"*80)
        print(xmlOutput)
        print("_"*80)
   
    def test08DeserializeAuthzDecisionQuery(self):
        samlUtil = SAMLUtil()
        authzDecisionQuery = samlUtil.buildAuthzDecisionQuery()
        
        # Create ElementTree Assertion Element
        authzDecisionQueryElem = AuthzDecisionQueryElementTree.toXML(
                                                            authzDecisionQuery)
        
        self.assert_(iselement(authzDecisionQueryElem))
        
        # Serialise to output 
        xmlOutput = prettyPrint(authzDecisionQueryElem)       
        self.assert_(len(xmlOutput))
        
        authzDecisionQueryStream = StringIO()
        authzDecisionQueryStream.write(xmlOutput)
        authzDecisionQueryStream.seek(0)

        tree = ElementTree.parse(authzDecisionQueryStream)
        elem2 = tree.getroot()
        
        authzDecisionQuery2 = AuthzDecisionQueryElementTree.fromXML(elem2)
        self.assert_(authzDecisionQuery2)
        self.assert_(
        authzDecisionQuery2.subject.nameID.value == SAMLTestCase.NAMEID_VALUE)
        self.assert_(
        authzDecisionQuery2.subject.nameID.format == SAMLTestCase.NAMEID_FORMAT)
        self.assert_(
            authzDecisionQuery2.issuer.value == SAMLTestCase.ISSUER_DN)
        self.assert_(
            authzDecisionQuery2.resource == SAMLTestCase.RESOURCE_URI)
        self.assert_(len(authzDecisionQuery2.actions) == 1)
        self.assert_(
            authzDecisionQuery2.actions[0].value == Action.HTTP_GET_ACTION)
        self.assert_(
            authzDecisionQuery2.actions[0].namespace == Action.GHPP_NS_URI)
        self.assert_(authzDecisionQuery2.evidence is None)

    def _createAuthzDecisionQueryResponse(self):
        """Helper method for Authz Decision Response"""
        response = Response()
        now = datetime.utcnow()
        response.issueInstant = now
        
        # Make up a request ID that this response is responding to
        response.inResponseTo = str(uuid4())
        response.id = str(uuid4())
        response.version = SAMLVersion(SAMLVersion.VERSION_20)
            
        response.issuer = Issuer()
        response.issuer.format = Issuer.X509_SUBJECT
        response.issuer.value = SAMLTestCase.ISSUER_DN
        
        response.status = Status()
        response.status.statusCode = StatusCode()
        response.status.statusCode.value = StatusCode.SUCCESS_URI
        response.status.statusMessage = StatusMessage()        
        response.status.statusMessage.value = "Response created successfully"
           
        assertion = Assertion()
        assertion.version = SAMLVersion(SAMLVersion.VERSION_20)
        assertion.id = str(uuid4())
        assertion.issueInstant = now
        
        authzDecisionStatement = AuthzDecisionStatement()
        authzDecisionStatement.decision = DecisionType.PERMIT
        authzDecisionStatement.resource = SAMLTestCase.RESOURCE_URI
        authzDecisionStatement.actions.append(Action())
        authzDecisionStatement.actions[-1].namespace = Action.GHPP_NS_URI
        authzDecisionStatement.actions[-1].value = Action.HTTP_GET_ACTION
        assertion.authzDecisionStatements.append(authzDecisionStatement)
        
        # Add a conditions statement for a validity of 8 hours
        assertion.conditions = Conditions()
        assertion.conditions.notBefore = now
        assertion.conditions.notOnOrAfter = now + timedelta(seconds=60*60*8)
               
        assertion.subject = Subject()  
        assertion.subject.nameID = NameID()
        assertion.subject.nameID.format = SAMLTestCase.NAMEID_FORMAT
        assertion.subject.nameID.value = SAMLTestCase.NAMEID_VALUE    
            
        assertion.issuer = Issuer()
        assertion.issuer.format = Issuer.X509_SUBJECT
        assertion.issuer.value = SAMLTestCase.ISSUER_DN

        response.assertions.append(assertion)
        
        return response
        
    def test09CreateAuthzDecisionQueryResponse(self):
        response = self._createAuthzDecisionQueryResponse()
        self.assert_(response.assertions[0])
        self.assert_(response.assertions[0].authzDecisionStatements[0])
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].decision == DecisionType.PERMIT)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].resource == SAMLTestCase.RESOURCE_URI)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].decision == DecisionType.PERMIT)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].actions[-1].namespace == Action.GHPP_NS_URI)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].actions[-1].value == Action.HTTP_GET_ACTION)
     
    def _serializeAuthzDecisionQueryResponse(self):
        response = self._createAuthzDecisionQueryResponse()
        
        # Create ElementTree Assertion Element
        responseElem = ResponseElementTree.toXML(response)
        self.assert_(iselement(responseElem))
        
        # Serialise to output        
        xmlOutput = prettyPrint(responseElem)
        return xmlOutput
    
    def test10SerializeAuthzDecisionQueryResponse(self):
        xmlOutput = self._serializeAuthzDecisionQueryResponse()
        self.assert_(len(xmlOutput))
        print("\n"+"_"*80)
        print(xmlOutput)
        print("_"*80)
        
        self.assert_('AuthzDecisionStatement' in xmlOutput)
        self.assert_('GET' in xmlOutput)
        self.assert_('Permit' in xmlOutput)

    def test11DeserializeAuthzDecisionResponse(self):
        xmlOutput = self._serializeAuthzDecisionQueryResponse()
        
        authzDecisionResponseStream = StringIO()
        authzDecisionResponseStream.write(xmlOutput)
        authzDecisionResponseStream.seek(0)

        tree = ElementTree.parse(authzDecisionResponseStream)
        elem = tree.getroot()
        response = ResponseElementTree.fromXML(elem)
        
        self.assert_(response.assertions[0])
        self.assert_(response.assertions[0].authzDecisionStatements[0])
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].decision == DecisionType.PERMIT)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].resource == SAMLTestCase.RESOURCE_URI)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].decision == DecisionType.PERMIT)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].actions[-1].namespace == Action.GHPP_NS_URI)
        self.assert_(response.assertions[0].authzDecisionStatements[0
            ].actions[-1].value == Action.HTTP_GET_ACTION)
        
 
    def test12PickleAssertion(self):
        # Test pickling with __slots__
        assertion = self._createAttributeAssertionHelper()
        assertion.issuer = Issuer()
        assertion.issuer.format = Issuer.X509_SUBJECT
        assertion.issuer.value = SAMLTestCase.ISSUER_DN
        
        jar = pickle.dumps(assertion)
        assertion2 = pickle.loads(jar)
        self.assert_(isinstance(assertion2, Assertion))
        self.assert_(assertion2.issuer.value == assertion.issuer.value)
        self.assert_(assertion2.issuer.format == assertion.issuer.format)
        self.assert_(len(assertion2.attributeStatements)==1)
        self.assert_(len(assertion2.attributeStatements[0].attributes) > 0)
        self.assert_(assertion2.attributeStatements[0].attributes[0
                     ].attributeValues[0
                     ].value == assertion.attributeStatements[0].attributes[0
                                ].attributeValues[0].value)
        
    def test13PickleAttributeQuery(self):
        # Test pickling with __slots__
        samlUtil = SAMLUtil()
        samlUtil.firstName = ''
        samlUtil.lastName = ''
        samlUtil.emailAddress = ''
        query = samlUtil.buildAttributeQuery(SAMLTestCase.ISSUER_DN,
                                             SAMLTestCase.NAMEID_VALUE)
        
        jar = pickle.dumps(query)
        query2 = pickle.loads(jar)

        self.assert_(isinstance(query2, AttributeQuery))
        self.assert_(query2.subject.nameID.value == query.subject.nameID.value)
        self.assert_((query2.subject.nameID.format == 
                      query.subject.nameID.format))
        self.assert_(query2.issuer.value == query.issuer.value)
        self.assert_(query2.issuer.format == query.issuer.format)
        self.assert_(query2.issueInstant == query.issueInstant)
        self.assert_(query2.id == query.id)
        self.assert_(len(query2.attributes) == 3)
        self.assert_(query2.attributes[0].name == "urn:esg:first:name")
        self.assert_(query2.attributes[1].nameFormat == SAMLUtil.XSSTRING_NS)

    def test14PickleAttributeQueryResponse(self):
        response = self._createAttributeQueryResponse()
        
        jar = pickle.dumps(response)
        response2 = pickle.loads(jar)
        
        self.assert_(isinstance(response2, Response))
        self.assert_((response2.status.statusCode.value == 
                      response.status.statusCode.value))
        self.assert_((response2.status.statusMessage.value == 
                      response.status.statusMessage.value))
        self.assert_(len(response2.assertions) == 1)
        self.assert_(response2.assertions[0].id == response.assertions[0].id)
        self.assert_((response2.assertions[0].conditions.notBefore == 
                      response.assertions[0].conditions.notBefore))
        self.assert_((response2.assertions[0].conditions.notOnOrAfter == 
                      response.assertions[0].conditions.notOnOrAfter))
        self.assert_(len(response2.assertions[0].attributeStatements) == 1)
        self.assert_(len(response2.assertions[0].attributeStatements[0
                                                            ].attributes) == 9)
        self.assert_(response2.assertions[0].attributeStatements[0].attributes[1
                     ].attributeValues[0
                     ].value == response.assertions[0].attributeStatements[0
                                    ].attributes[1].attributeValues[0].value)
             
    def test15PickleAuthzDecisionQuery(self):
        samlUtil = SAMLUtil()
        query = samlUtil.buildAuthzDecisionQuery()
             
        jar = pickle.dumps(query)
        query2 = pickle.loads(jar)
        
        self.assert_(isinstance(query2, AuthzDecisionQuery))
        self.assert_(query.resource == query2.resource)
        self.assert_(query.version == query2.version)
        self.assert_(len(query2.actions) == 1)
        self.assert_(query2.actions[0].value == Action.HTTP_GET_ACTION)
        self.assert_(query2.actions[0].namespace == Action.GHPP_NS_URI)

    def test16PickleAuthzDecisionResponse(self):
        response = self._createAuthzDecisionQueryResponse()
        
        jar = pickle.dumps(response)
        response2 = pickle.loads(jar)
        
        self.assert_(isinstance(response2, Response))
        
        self.assert_(len(response.assertions) == 1)
        self.assert_(len(response.assertions[0].authzDecisionStatements) == 1)
         
        self.assert_(response.assertions[0].authzDecisionStatements[0
                        ].resource == response2.assertions[0
                                        ].authzDecisionStatements[0].resource)
        
        self.assert_(len(response.assertions[0].authzDecisionStatements[0
                        ].actions) == 1)
        self.assert_(response.assertions[0].authzDecisionStatements[0
                        ].actions[0].value == response2.assertions[0
                                        ].authzDecisionStatements[0
                                                ].actions[0].value)
        
        self.assert_(response2.assertions[0].authzDecisionStatements[0
                        ].actions[0].namespace == Action.GHPP_NS_URI)        

        self.assert_(response2.assertions[0].authzDecisionStatements[0
                        ].decision == DecisionType.PERMIT)        
        
if __name__ == "__main__":
    unittest.main()        



"""SAML 2.0 bindings module implements SOAP binding for Authorisation Decision
query

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "12/02/10"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: authzdecisionquery.py 7130 2010-06-30 13:33:07Z pjkersha $'
import logging
log = logging.getLogger(__name__)

from M2Crypto.m2urllib2 import HTTPSHandler

from ndg.saml.saml2.core import AuthzDecisionQuery

from ndg.saml.saml2.binding.soap.client.subjectquery import (
                                                    SubjectQuerySOAPBinding,
                                                    SubjectQueryResponseError)

# Prevent whole module breaking if this is not available - it's only needed for
# AuthzDecisionQuerySslSOAPBinding
try:
    from ndg.saml.utils.m2crypto import SSLContextProxy
    _sslContextProxySupport = True
    
except ImportError:
    _sslContextProxySupport = False
    

class AuthzDecisionQueryResponseError(SubjectQueryResponseError):
    """SAML Response error from Attribute Query"""
    

class AuthzDecisionQuerySOAPBinding(SubjectQuerySOAPBinding): 
    """SAML Attribute Query SOAP Binding
    
    Nb. Assumes X.509 subject type for query issuer
    """
    SERIALISE_KW = 'serialise'
    DESERIALISE_KW = 'deserialise'
    QUERY_TYPE = AuthzDecisionQuery
    __slots__ = ('__resourceURI', '__action', '__actionNs')
    
    def __init__(self, **kw):
        '''Create SOAP Client for SAML Authorization Decision Query'''
        cls = AuthzDecisionQuerySOAPBinding
        
        # Default to ElementTree based serialisation/deserialisation
        if cls.SERIALISE_KW not in kw:
            from ndg.saml.xml.etree import AuthzDecisionQueryElementTree
            kw[cls.SERIALISE_KW] = AuthzDecisionQueryElementTree.toXML
               
        if cls.DESERIALISE_KW not in kw:
            from ndg.saml.xml.etree import ResponseElementTree
            kw[cls.DESERIALISE_KW] = ResponseElementTree.fromXML

        super(AuthzDecisionQuerySOAPBinding, self).__init__(**kw)

    def _getResourceURI(self):
        return self.query.resource

    def _setResourceURI(self, value):
        self.query.resource = value
        
    resourceURI = property(_getResourceURI, _setResourceURI, 
                           doc="Resource URI to query for access")
    
    @property
    def actions(self):
        return self.query.actions

    
class AuthzDecisionQuerySslSOAPBinding(AuthzDecisionQuerySOAPBinding):
    """Specialisation of AuthzDecisionQuerySOAPbinding taking in the setting of
    SSL parameters for mutual authentication
    """
    SSL_CONTEXT_PROXY_SUPPORT = _sslContextProxySupport
    __slots__ = ('__sslCtxProxy',)
    
    def __init__(self, **kw):
        if not AuthzDecisionQuerySslSOAPBinding.SSL_CONTEXT_PROXY_SUPPORT:
            raise ImportError("ndg.security.common.utils.m2crypto import "
                              "failed - missing M2Crypto package?")
        
        # Miss out default HTTPSHandler and set in send() instead
        if 'handlers' in kw:
            raise TypeError("__init__() got an unexpected keyword argument "
                            "'handlers'")
            
        super(AuthzDecisionQuerySslSOAPBinding, self).__init__(handlers=(), 
                                                               **kw)
        self.__sslCtxProxy = SSLContextProxy()

    def send(self, **kw):
        """Override base class implementation to pass explicit SSL Context
        """
        httpsHandler = HTTPSHandler(ssl_context=self.sslCtxProxy.createCtx())
        self.client.openerDirector.add_handler(httpsHandler)
        return super(AuthzDecisionQuerySslSOAPBinding, self).send(**kw)
        
    @property
    def sslCtxProxy(self):
        """SSL Context Proxy object used for setting up an SSL Context for
        queries
        """
        return self.__sslCtxProxy
            
    def __setattr__(self, name, value):
        """Enable setting of SSLContextProxy attributes as if they were 
        attributes of this class.  This is intended as a convenience for 
        making settings parameters read from a config file
        """
        try:
            super(AuthzDecisionQuerySslSOAPBinding, self).__setattr__(name, 
                                                                      value)
            
        except AttributeError, e:
            # Coerce into setting SSL Context Proxy attributes
            try:
                setattr(self.sslCtxProxy, name, value)
            except:
                raise e



"""NDG SAML

WSGI Package for SAML SOAP Binding 

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "30/06/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7142 2010-06-30 15:39:36Z pjkersha $'
